@@ -51,7 +51,6 @@ extern int key_setnet( struct key_netstarg *arg );
 #endif
 #ifdef HAVE_SETUSERCONTEXT
 # include <login_cap.h>
-# define USE_LOGIN_CAP 1
 #endif
 #ifdef USE_PAM
 # ifdef HAVE_PAM_PAM_APPL_H
@@ -66,9 +65,6 @@ extern int loginrestrictions( const char *Name, const int Mode, const char *Tty,
 extern int loginfailed( const char *User, const char *Host, const char *Tty );
 extern int loginsuccess( const char *User, const char *Host, const char *Tty, char **Msg );
 #else /* USE_PAM || _AIX */
-# ifdef USESHADOW
-#  include <shadow.h>
-# endif
 # ifdef KERBEROS
 #  include <sys/param.h>
 #  include <krb.h>
@@ -82,6 +78,9 @@ extern int loginsuccess( const char *User, const char *Host, const char *Tty, ch
 /* for expiration */
 # include <time.h>
 #endif /* USE_PAM || _AIX */
+#ifdef HAVE_SHADOW
+# include <shadow.h>
+#endif
 
 /*
  * Session data, mostly what struct verify_info was for
@@ -101,6 +100,13 @@ char *curdmrc;
 char *newdmrc;
 
 static struct passwd *p;
+#ifdef HAVE_SETUSERCONTEXT
+# ifdef HAVE_LOGIN_GETCLASS
+login_cap_t *lc;
+# else
+struct login_cap *lc;
+# endif
+#endif
 #ifdef USE_PAM
 static pam_handle_t *pamh;
 #elif defined(_AIX)
@@ -291,50 +297,32 @@ doPAMAuth( const char *psrv, struct pam_data *pdata )
 	pconv.conv = PAM_conv;
 	pconv.appdata_ptr = (void *)pdata;
 	Debug( " PAM service %s\n", psrv );
-	if (pamh) {
-		pam_get_item( pamh, PAM_SERVICE, &pitem );
+	if ((pretc = pam_start( psrv, curuser, &pconv, &pamh )) != PAM_SUCCESS)
+		goto pam_bail2;
+	if ((pretc = pam_set_item( pamh, PAM_TTY, td->name )) != PAM_SUCCESS) {
+	  pam_bail:
+		pam_end( pamh, pretc );
+		pamh = 0;
+	  pam_bail2:
 		ReInitErrorLog();
-		if (strcmp( (const char *)pitem, psrv )) {
-			Debug( "closing old PAM handle\n" );
-			pam_end( pamh, PAM_SUCCESS );
-			ReInitErrorLog();
-			goto opennew;
-		}
-		Debug( "reusing old PAM handle\n" );
-/*  this makes linux-pam crash ...
-		if ((pretc = pam_set_item (pamh, PAM_SERVICE, psrv)) != PAM_SUCCESS)
-			goto pam_bail;
-*/
-		if ((pretc = pam_set_item( pamh, PAM_CONV, &pconv )) != PAM_SUCCESS) {
-		  pam_bail:
-			pam_end( pamh, pretc );
-			pamh = 0;
-		  pam_bail2:
-			ReInitErrorLog();
-			LogError( "PAM error: %s\n", pam_strerror( pamh, pretc ) );
-			V_RET_FAIL( 0 );
-		}
-		if ((pretc = pam_set_item( pamh, PAM_USER, curuser )) != PAM_SUCCESS)
-			goto pam_bail;
-	} else {
-	  opennew:
-		Debug( "opening new PAM handle\n" );
-		if ((pretc = pam_start( psrv, curuser, &pconv, &pamh )) != PAM_SUCCESS)
-			goto pam_bail2;
-		if ((pretc = pam_set_item( pamh, PAM_TTY, td->name )) != PAM_SUCCESS)
-			goto pam_bail;
-		if ((td->displayType & d_location) == dForeign) {
-			char *cp = strchr( td->name, ':' );
-			*cp = 0;
-			pretc = pam_set_item( pamh, PAM_RHOST, td->name );
-			*cp = ':';
-			if (pretc != PAM_SUCCESS)
-				goto pam_bail;
-		}
-# ifdef PAM_FAIL_DELAY
-		pam_set_item( pamh, PAM_FAIL_DELAY, (void *)fail_delay );
-# endif
+		LogError( "PAM error: %s\n", pam_strerror( 0, pretc ) );
+		V_RET_FAIL( 0 );
 	}
+	if ((td->displayType & d_location) == dForeign) {
+		char *cp = strchr( td->name, ':' );
+		*cp = 0;
+		pretc = pam_set_item( pamh, PAM_RHOST, td->name );
+		*cp = ':';
+		if (pretc != PAM_SUCCESS)
+			goto pam_bail;
+	}
+# ifdef __sun__ /* Only Solaris <= 9, but checking it does not seem worth it. */
+	else if (pam_set_item( pamh, PAM_RHOST, 0 ) != PAM_SUCCESS)
+		goto pam_bail;
+# endif
+# ifdef PAM_FAIL_DELAY
+	pam_set_item( pamh, PAM_FAIL_DELAY, (void *)fail_delay );
+# endif
 	ReInitErrorLog();
 
 	Debug( " pam_authenticate() ...\n" );
@@ -342,8 +330,11 @@ doPAMAuth( const char *psrv, struct pam_data *pdata )
 	                          td->allowNullPasswd ? 0 : PAM_DISALLOW_NULL_AUTHTOK );
 	ReInitErrorLog();
 	Debug( " pam_authenticate() returned: %s\n", pam_strerror( pamh, pretc ) );
-	if (pdata->abort)
+	if (pdata->abort) {
+		pam_end( pamh, PAM_SUCCESS );
+		pamh = 0;
 		return 0;
+	}
 	if (!curuser) {
 		Debug( " asking PAM for user ...\n" );
 		pam_get_item( pamh, PAM_USER, &pitem );
@@ -358,8 +349,12 @@ doPAMAuth( const char *psrv, struct pam_data *pdata )
 		case PAM_AUTH_ERR:
 		case PAM_MAXTRIES: /* should handle this better ... */
 		case PAM_AUTHINFO_UNAVAIL: /* returned for unknown users ... bogus */
+			pam_end( pamh, pretc );
+			pamh = 0;
 			V_RET_AUTH;
 		default:
+			pam_end( pamh, pretc );
+			pamh = 0;
 			V_RET_FAIL( 0 );
 		}
 	}
@@ -373,6 +368,9 @@ static int
 AccNoPass( const char *un )
 {
 	struct passwd *pw = 0;
+# ifdef HAVE_SHADOW /* (sic!) - not USESHADOW */
+	struct spwd *spw;
+# endif
 #else
 AccNoPass( const char *un, struct passwd *pw )
 {
@@ -396,6 +394,13 @@ AccNoPass( const char *un, struct passwd *pw )
 #if defined(USE_PAM) || defined(_AIX)
 			if (!(pw = getpwnam( un )))
 				return 0;
+			if (pw->pw_passwd[0] == '!' || pw->pw_passwd[0] == '*')
+				continue;
+# ifdef HAVE_SHADOW /* (sic!) - not USESHADOW */
+			if ((spw = getspnam( un )) &&
+			    (spw->sp_pwdp[0] == '!' || spw->sp_pwdp[0] == '*'))
+					continue;
+# endif
 #endif
 			if (pw->pw_uid)
 				return 1;
@@ -425,7 +430,7 @@ AccNoPass( const char *un, struct passwd *pw )
 	return 0;
 }
 
-#if !defined(USE_PAM) && !defined(_AIX) && defined(USE_LOGIN_CAP)
+#if !defined(USE_PAM) && !defined(_AIX) && defined(HAVE_SETUSERCONTEXT)
 # define LC_RET0 do { login_close(lc); return 0; } while(0)
 #else
 # define LC_RET0 return 0
@@ -453,13 +458,6 @@ Verify( GConvFunc gconv, int rootok )
 # if defined(HAVE_STRUCT_PASSWD_PW_EXPIRE) || defined(USESHADOW)
 	int tim, expir, warntime, quietlog;
 # endif
-# ifdef USE_LOGIN_CAP
-#  ifdef HAVE_LOGIN_GETCLASS
-	login_cap_t *lc;
-#  else
-	struct login_cap *lc;
-#  endif
-# endif
 #endif
 
 	Debug( "Verify ...\n" );
@@ -598,25 +596,31 @@ Verify( GConvFunc gconv, int rootok )
 
 	if (!(p = getpwnam( curuser ))) {
 		Debug( "getpwnam() failed.\n" );
+		gconv( GCONV_PASS, 0 );
 		V_RET_AUTH;
 	}
-# ifdef __linux__ /* only Linux? */
 	if (p->pw_passwd[0] == '!' || p->pw_passwd[0] == '*') {
 		Debug( "account is locked\n" );
+		gconv( GCONV_PASS, 0 );
 		V_RET_AUTH;
 	}
-# endif
 
 # ifdef USESHADOW
-	if ((sp = getspnam( curuser )))
+	if ((sp = getspnam( curuser ))) {
 		p->pw_passwd = sp->sp_pwdp;
-	else
+		if (p->pw_passwd[0] == '!' || p->pw_passwd[0] == '*') {
+			Debug( "account is locked\n" );
+			gconv( GCONV_PASS, 0 );
+			V_RET_AUTH;
+		}
+	} else
 		Debug( "getspnam() failed: %m. Are you root?\n" );
 # endif
 
 	if (!*p->pw_passwd) {
 		if (!td->allowNullPasswd) {
 			Debug( "denying user with empty password\n" );
+			gconv( GCONV_PASS, 0 );
 			V_RET_AUTH;
 		}
 		goto nplogin;
@@ -704,6 +708,8 @@ Verify( GConvFunc gconv, int rootok )
 		pdata.gconv = conv_interact;
 		/* pam will have output a message already, so no PrepErrorGreet () */
 		if (gconv != conv_interact || pnopass) {
+			pam_end( pamh, PAM_SUCCESS );
+			pamh = 0;
 			GSendInt( V_CHTOK_AUTH );
 			/* this cannot auth the wrong user, as only classic auths get here */
 			while (!doPAMAuth( PAMService, &pdata ))
@@ -717,8 +723,11 @@ Verify( GConvFunc gconv, int rootok )
 			pretc = pam_chauthtok( pamh, PAM_CHANGE_EXPIRED_AUTHTOK );
 			ReInitErrorLog();
 			Debug( " pam_chauthtok() returned: %s\n", pam_strerror( pamh, pretc ) );
-			if (pdata.abort)
+			if (pdata.abort) {
+				pam_end( pamh, PAM_SUCCESS );
+				pamh = 0;
 				return 0;
+			}
 			if (pretc == PAM_SUCCESS)
 				break;
 			/* effectively there is only PAM_AUTHTOK_ERR */
@@ -728,8 +737,11 @@ Verify( GConvFunc gconv, int rootok )
 			free( curpass );
 		curpass = newpass;
 		newpass = 0;
-	} else if (pretc != PAM_SUCCESS)
+	} else if (pretc != PAM_SUCCESS) {
+		pam_end( pamh, pretc );
+		pamh = 0;
 		V_RET_AUTH;
+	}
 
 #elif defined(_AIX) /* USE_PAM */
 
@@ -751,23 +763,11 @@ Verify( GConvFunc gconv, int rootok )
 	if (msg)
 		free( (void *)msg );
 
-#else /* USE_PAM || _AIX */
+#endif /* USE_PAM || _AIX */
 
-# ifdef HAVE_GETUSERSHELL
-	for (;;) {
-		if (!(s = getusershell())) {
-			Debug( "shell not in /etc/shells\n" );
-			endusershell();
-			V_RET_FAIL( "Your login shell is not listed in /etc/shells" );
-		}
-		if (!strcmp( s, p->pw_shell )) {
-			endusershell();
-			break;
-		}
-	}
-# endif
+#ifndef _AIX
 
-# ifdef USE_LOGIN_CAP
+# ifdef HAVE_SETUSERCONTEXT
 #  ifdef HAVE_LOGIN_GETCLASS
 	lc = login_getclass( p->pw_class );
 #  else
@@ -775,106 +775,45 @@ Verify( GConvFunc gconv, int rootok )
 #  endif
 	if (!lc)
 		V_RET_FAIL( 0 );
+
+	p->pw_shell = login_getcapstr( lc, "shell", p->pw_shell, p->pw_shell );
 # endif
 
+# ifndef USE_PAM
 
-/* restrict_nologin */
-# ifndef _PATH_NOLOGIN
-#  define _PATH_NOLOGIN "/etc/nologin"
-# endif
+/* restrict_expired */
+#  if defined(HAVE_STRUCT_PASSWD_PW_EXPIRE) || defined(USESHADOW)
 
-	if ((
-# ifdef USE_LOGIN_CAP
-	     /* Do we ignore a nologin file? */
-	     !login_getcapbool( lc, "ignorenologin", 0 )) &&
-	    (!stat( (nolg = login_getcapstr( lc, "nologin", "", NULL )), &st ) ||
-# endif
-		 !stat( (nolg = _PATH_NOLOGIN), &st )))
-	{
-		PrepErrorGreet();
-		GSendInt( V_MSG_ERR );
-		if (st.st_size && (fd = open( nolg, O_RDONLY )) >= 0) {
-			if ((buf = Malloc( st.st_size + 1 ))) {
-				if (read( fd, buf, st.st_size ) == st.st_size) {
-					buf[st.st_size] = 0;
-					GSendStr( buf );
-					free( buf );
-					close( fd );
-					GSendInt( V_FAIL );
-					LC_RET0;
-				}
-				free( buf );
-			}
-			close( fd );
-		}
-		GSendStr( "Logins are not allowed at the moment.\nTry again later" );
-		GSendInt( V_FAIL );
-		LC_RET0;
-	}
-
-
-/* restrict_nohome */
-# ifdef USE_LOGIN_CAP
-	if (login_getcapbool( lc, "requirehome", 0 )) {
-		struct stat st;
-		if (!*p->pw_dir || stat( p->pw_dir, &st ) || st.st_uid != p->pw_uid) {
-			PrepErrorGreet();
-			GSendInt( V_MSG_ERR );
-			GSendStr( "Home folder not available" );
-			GSendInt( V_FAIL );
-			LC_RET0;
-		}
-	}
-# endif
-
-
-/* restrict_time */
-# ifdef USE_LOGIN_CAP
-#  ifdef HAVE_AUTH_TIMEOK
-	if (!auth_timeok( lc, time( NULL ) )) {
-		PrepErrorGreet();
-		GSendInt( V_MSG_ERR );
-		GSendStr( "You are not allowed to login at the moment" );
-		GSendInt( V_FAIL );
-		LC_RET0;
-	}
-#  endif
-# endif
-
-
-/* restrict_expired; this MUST be the last one */
-# if defined(HAVE_STRUCT_PASSWD_PW_EXPIRE) || defined(USESHADOW)
-
-#  if !defined(HAVE_STRUCT_PASSWD_PW_EXPIRE) || (!defined(USE_LOGIN_CAP) && defined(USESHADOW))
+#   if !defined(HAVE_STRUCT_PASSWD_PW_EXPIRE) || (!defined(HAVE_SETUSERCONTEXT) && defined(USESHADOW))
 	if (sp)
-#  endif
+#   endif
 	{
 
-#  define DEFAULT_WARN	(2L * 7L)  /* Two weeks */
+#   define DEFAULT_WARN	(2L * 7L)  /* Two weeks */
 
 		tim = time( NULL ) / 86400L;
 
-#  ifdef USE_LOGIN_CAP
+#   ifdef HAVE_SETUSERCONTEXT
 		quietlog = login_getcapbool( lc, "hushlogin", 0 );
 		warntime = login_getcaptime( lc, "warnexpire",
 		                             DEFAULT_WARN * 86400L,
 		                             DEFAULT_WARN * 86400L ) / 86400L;
-#  else
+#   else
 		quietlog = 0;
-#	ifdef USESHADOW
+#    ifdef USESHADOW
 		warntime = sp->sp_warn != -1 ? sp->sp_warn : DEFAULT_WARN;
-#	else
+#    else
 		warntime = DEFAULT_WARN;
-#	endif
-#  endif
+#    endif
+#   endif
 
-#  ifdef HAVE_STRUCT_PASSWD_PW_EXPIRE
+#   ifdef HAVE_STRUCT_PASSWD_PW_EXPIRE
 		if (p->pw_expire) {
 			expir = p->pw_expire / 86400L;
-#  else
+#   else
 		if (sp->sp_expire != -1) {
 			expir = sp->sp_expire;
-#  endif
+#   endif
 			if (tim > expir) {
 				PrepErrorGreet();
 				GSendInt( V_MSG_ERR );
@@ -895,10 +834,10 @@ Verify( GConvFunc gconv, int rootok )
 			}
 		}
 
-#  ifdef HAVE_STRUCT_PASSWD_PW_EXPIRE
+#   ifdef HAVE_STRUCT_PASSWD_PW_EXPIRE
 		if (p->pw_change) {
 			expir = p->pw_change / 86400L;
-#  else
+#   else
 		if (!sp->sp_lstchg) {
 			PrepErrorGreet();
 			GSendInt( V_MSG_ERR );
@@ -917,7 +856,7 @@ Verify( GConvFunc gconv, int rootok )
 				GSendInt( V_FAIL );
 				LC_RET0;
 			}
-#  endif
+#   endif
 			if (tim > expir) {
 				PrepErrorGreet();
 				GSendInt( V_MSG_ERR );
@@ -941,13 +880,84 @@ Verify( GConvFunc gconv, int rootok )
 
 	}
 
-# endif /* HAVE_STRUCT_PASSWD_PW_EXPIRE || USESHADOW */
+#  endif /* HAVE_STRUCT_PASSWD_PW_EXPIRE || USESHADOW */
 
-# ifdef USE_LOGIN_CAP
-	login_close( lc );
+/* restrict_nologin */
+#  ifndef _PATH_NOLOGIN
+#   define _PATH_NOLOGIN "/etc/nologin"
+#  endif
+
+	if ((
+#  ifdef HAVE_SETUSERCONTEXT
+	     /* Do we ignore a nologin file? */
+	     !login_getcapbool( lc, "ignorenologin", 0 )) &&
+	    (!stat( (nolg = login_getcapstr( lc, "nologin", "", NULL )), &st ) ||
+#  endif
+		 !stat( (nolg = _PATH_NOLOGIN), &st )))
+	{
+		PrepErrorGreet();
+		GSendInt( V_MSG_ERR );
+		if (st.st_size && (fd = open( nolg, O_RDONLY )) >= 0) {
+			if ((buf = Malloc( st.st_size + 1 ))) {
+				if (read( fd, buf, st.st_size ) == st.st_size) {
+					buf[st.st_size] = 0;
+					GSendStr( buf );
+					free( buf );
+					close( fd );
+					GSendInt( V_FAIL );
+					LC_RET0;
+				}
+				free( buf );
+			}
+			close( fd );
+		}
+		GSendStr( "Logins are not allowed at the moment.\nTry again later" );
+		GSendInt( V_FAIL );
+		LC_RET0;
+	}
+
+/* restrict_time */
+#  if defined(HAVE_SETUSERCONTEXT) && defined(HAVE_AUTH_TIMEOK)
+	if (!auth_timeok( lc, time( NULL ) )) {
+		PrepErrorGreet();
+		GSendInt( V_MSG_ERR );
+		GSendStr( "You are not allowed to login at the moment" );
+		GSendInt( V_FAIL );
+		LC_RET0;
+	}
+#  endif
+
+#  ifdef HAVE_GETUSERSHELL
+	for (;;) {
+		if (!(s = getusershell())) {
+			Debug( "shell not in /etc/shells\n" );
+			endusershell();
+			V_RET_FAIL( "Your login shell is not listed in /etc/shells" );
+		}
+		if (!strcmp( s, p->pw_shell )) {
+			endusershell();
+			break;
+		}
+	}
+#  endif
+
+# endif /* !USE_PAM */
+
+/* restrict_nohome */
+# ifdef HAVE_SETUSERCONTEXT
+	if (login_getcapbool( lc, "requirehome", 0 )) {
+		struct stat st;
+		if (!*p->pw_dir || stat( p->pw_dir, &st ) || st.st_uid != p->pw_uid) {
+			PrepErrorGreet();
+			GSendInt( V_MSG_ERR );
+			GSendStr( "Home folder not available" );
+			GSendInt( V_FAIL );
+			LC_RET0;
+		}
+	}
 # endif
 
-#endif /* USE_PAM || _AIX */
+#endif /* !_AIX */
 
 	return 1;
 
@@ -963,6 +973,54 @@ static const char *envvars[] = {
 };
 
 
+#if defined(USE_PAM) && defined(HAVE_INITGROUPS)
+static int num_saved_gids;
+static gid_t *saved_gids;
+
+static int
+saveGids( void )
+{
+	num_saved_gids = getgroups( 0, 0 );
+	if (!(saved_gids = Malloc( sizeof(gid_t) * num_saved_gids )))
+		return 0;
+	if (getgroups( num_saved_gids, saved_gids ) < 0) {
+		LogError( "saving groups failed: %m\n" );
+		return 0;
+	}
+	return 1;
+}
+
+static int
+restoreGids( void )
+{
+	if (setgroups( num_saved_gids, saved_gids ) < 0) {
+		LogError( "restoring groups failed: %m\n" );
+		return 0;
+	}
+	if (setgid( p->pw_gid ) < 0) {
+		LogError( "restoring gid failed: %m\n" );
+		return 0;
+	}
+	return 1;
+}
+#endif /* USE_PAM && HAVE_INITGROUPS */
+
+static int
+resetGids( void )
+{
+#ifdef HAVE_INITGROUPS
+	if (setgroups( 0, &p->pw_gid /* anything */ ) < 0) {
+		LogError( "restoring groups failed: %m\n" );
+		return 0;
+	}
+#endif
+	if (setgid( 0 ) < 0) {
+		LogError( "restoring gid failed: %m\n" );
+		return 0;
+	}
+	return 1;
+}
+
 static int
 SetGid( const char *name, int gid )
 {
@@ -973,6 +1031,7 @@ SetGid( const char *name, int gid )
 #ifdef HAVE_INITGROUPS
 	if (initgroups( name, gid ) < 0) {
 		LogError( "initgroups for %s failed: %m\n", name );
+		setgid( 0 );
 		return 0;
 	}
 #endif	 /* QNX4 doesn't support multi-groups, no initgroups() */
@@ -992,7 +1051,12 @@ SetUid( const char *name, int uid )
 static int
 SetUser( const char *name, int uid, int gid )
 {
-	return SetGid( name, gid ) && SetUid( name, uid );
+	if (SetGid( name, gid )) {
+		if (SetUid( name, uid ))
+			return 1;
+		resetGids();
+	}
+	return 0;
 }
 
 #if defined(SECURE_RPC) || defined(K5AUTH)
@@ -1051,6 +1115,10 @@ mergeSessionArgs( int cansave )
 }
 
 static int removeAuth;
+#ifdef USE_PAM
+static int removeSession;
+static int removeCreds;
+#endif
 
 int
 StartClient()
@@ -1060,6 +1128,9 @@ StartClient()
 	char **argv, *fname, *str;
 #ifdef USE_PAM
 	char **pam_env;
+# ifdef _AIX
+	char **saved_env;
+# endif
 	struct pam_conv pconv;
 	int pretc;
 #else
@@ -1147,7 +1218,7 @@ StartClient()
 		env = setEnv( env, "KRBTKFILE", krbtkfile );
 #endif
 	userEnviron = inheritEnv( env, envvars );
-	env = systemEnv( curuser );
+	env = systemEnv( p->pw_name );
 	systemEnviron = setEnv( env, "HOME", p->pw_dir );
 	Debug( "user environment:\n%[|''>'\n's"
 	       "system environment:\n%[|''>'\n's"
@@ -1197,13 +1268,83 @@ StartClient()
 		mergeSessionArgs( TRUE );
 
 	Debug( "now starting the session\n" );
+
 #ifdef USE_PAM
+	/* the greeter is gone by now ... */
 	pconv.conv = PAM_conv_null;
 	pconv.appdata_ptr = 0;
-	pam_set_item( pamh, PAM_CONV, &pconv ); /* XXX this can fail */
-	pam_open_session( pamh, 0 ); /* XXX this can fail, too */
+	if ((pretc = pam_set_item( pamh, PAM_CONV, &pconv )) != PAM_SUCCESS) {
+		ReInitErrorLog();
+		LogError( "pam_set_item() for %s failed: %s\n",
+		          curuser, pam_strerror( pamh, pretc ) );
+		return 0;
+	}
 	ReInitErrorLog();
 #endif
+
+#ifdef USE_PAM
+
+# ifdef HAVE_SETUSERCONTEXT
+	if (setusercontext( lc, p, p->pw_uid, LOGIN_SETGROUP )) {
+		LogError( "setusercontext(groups) for %s failed: %m\n",
+		          curuser );
+		return 0;
+	}
+# else
+	if (!SetGid( curuser, curgid ))
+		return 0;
+# endif
+
+# ifdef _AIX
+	if (!(pam_env = initStrArr( 0 ))) {
+		resetGids();
+		return 0;
+	}
+	saved_env = environ;
+	environ = pam_env;
+# endif
+	removeCreds = 1; /* set it first - i don't trust PAM's rollback */
+	pretc = pam_setcred( pamh, 0 );
+	ReInitErrorLog();
+# ifdef _AIX
+	pam_env = environ;
+	environ = saved_env;
+# endif
+# ifdef HAVE_INITGROUPS
+	/* This seems to be a strange place for it, but do it:
+	   - after the initial groups are set
+	   - after pam_setcred might have set something, even in the error case
+	   - before pam_setcred(DELETE_CRED) might need it
+	 */
+	if (!saveGids())
+		return 0;
+# endif
+	if (pretc != PAM_SUCCESS) {
+		LogError( "pam_setcred() for %s failed: %s\n",
+		          curuser, pam_strerror( pamh, pretc ) );
+		resetGids();
+		return 0;
+	}
+
+	removeSession = 1; /* set it first - same as above */
+	pretc = pam_open_session( pamh, 0 );
+	ReInitErrorLog();
+	if (pretc != PAM_SUCCESS) {
+		LogError( "pam_open_session() for %s failed: %s\n",
+		          curuser, pam_strerror( pamh, pretc ) );
+		resetGids();
+		return 0;
+	}
+
+	/* we don't want sessreg and the startup/reset scripts run with user
+	   credentials. unfortunately, we can reset only the gids. */
+	resetGids();
+
+# define D_LOGIN_SETGROUP LOGIN_SETGROUP
+#else /* USE_PAM */
+# define D_LOGIN_SETGROUP 0
+#endif /* USE_PAM */
+
 	removeAuth = 1;
 	chownCtrl( &td->ctrl, curuid );
 	endpwent();
@@ -1215,6 +1356,7 @@ StartClient()
 	case 0:
 
 		sessreg( td, getpid(), curuser, curuid );
+
 		if (source( systemEnviron, td->startup, td_setup )) {
 			LogError( "Cannot execute startup script %\"s\n", td->startup );
 			exit( 1 );
@@ -1224,32 +1366,15 @@ StartClient()
 			exit( 1 );
 		GSet( &mstrtalk );
 
-		/* Do system-dependent login setup here */
 		setsid();
 
 	/* Memory leaks are ok here as we exec() soon. */
 
 #if defined(USE_PAM) || !defined(_AIX)
 
-# ifndef HAVE_SETUSERCONTEXT
-		if (!SetGid( curuser, curgid ))
-			exit( 1 );
-# endif
 # ifdef USE_PAM
-#  ifdef _AIX
-		if (!(environ = initStrArr( 0 )))
-			exit( 1 );
-#  endif
-		if ((pretc = pam_setcred( pamh, 0 )) != PAM_SUCCESS) {
-			ReInitErrorLog();
-			LogError( "pam_setcred() for %s failed: %s\n",
-			          curuser, pam_strerror( pamh, pretc ) );
-			exit( 1 );
-		}
 		/* pass in environment variables set by libpam and modules it called */
-#  ifdef _AIX
-		pam_env = environ;
-#  else
+#  ifndef _AIX
 		pam_env = pam_getenvlist( pamh );
 		ReInitErrorLog();
 #  endif
@@ -1257,19 +1382,36 @@ StartClient()
 			for (; *pam_env; pam_env++)
 				userEnviron = putEnv( *pam_env, userEnviron );
 # endif
-# ifndef HAVE_SETUSERCONTEXT
-#  ifdef HAVE_SETLOGIN
+
+# ifdef HAVE_SETLOGIN
 		if (setlogin( curuser ) < 0) {
 			LogError( "setlogin for %s failed: %m\n", curuser );
 			exit( 1 );
 		}
-#  endif
+#  define D_LOGIN_SETLOGIN LOGIN_SETLOGIN
+# else
+#  define D_LOGIN_SETLOGIN 0
+# endif
+
+# if defined(USE_PAM) && defined(HAVE_INITGROUPS)
+		if (!restoreGids())
+			exit( 1 );
+# endif
+
+# ifndef HAVE_SETUSERCONTEXT
+
+#  ifdef USE_PAM
 		if (!SetUid( curuser, curuid ))
 			exit( 1 );
-# else /* HAVE_SETUSERCONTEXT */
+#  else
+		if (!SetUser( curuser, curuid, curgid ))
+			exit( 1 );
+#  endif
+
+# else /* !HAVE_SETUSERCONTEXT */
 
 		/*
-		 * Destroy environment unless user has requested its preservation.
+		 * Destroy environment.
 		 * We need to do this before setusercontext() because that may
 		 * set or reset some environment variables.
 		 */
@@ -1280,7 +1422,9 @@ StartClient()
 		 * Set the user's credentials: uid, gid, groups,
 		 * environment variables, resource limits, and umask.
 		 */
-		if (setusercontext( NULL, p, p->pw_uid, LOGIN_SETALL ) < 0) {
+		if (setusercontext( lc, p, p->pw_uid,
+		        LOGIN_SETALL & ~(D_LOGIN_SETGROUP|D_LOGIN_SETLOGIN) ) < 0)
+		{
 			LogError( "setusercontext for %s failed: %m\n", curuser );
 			exit( 1 );
 		}
@@ -1288,8 +1432,9 @@ StartClient()
 		for (i = 0; environ[i]; i++)
 			userEnviron = putEnv( environ[i], userEnviron );
 
-# endif /* HAVE_SETUSERCONTEXT */
-#else /* _AIX */
+# endif /* !HAVE_SETUSERCONTEXT */
+
+#else /* PAM || !_AIX */
 		/*
 		 * Set the user's credentials: uid, gid, groups,
 		 * audit classes, user limits, and umask.
@@ -1392,7 +1537,7 @@ StartClient()
 			}
 			ASPrintf( &lname, td->clientLogFile, td->name );
 			if ((lfd = creat( lname, 0600 )) < 0) {
-				LogInfo( "Cannot create session log file %s: %m\n", lname );
+				LogWarn( "Cannot create session log file %s: %m\n", lname );
 				free( lname );
 				goto tmperr;
 			}
@@ -1473,56 +1618,77 @@ StartClient()
 void
 SessionExit( int status )
 {
+	int pid;
+#ifdef USE_PAM
+	int pretc;
+#endif
+
+	if (removeAuth) {
+		if (source( systemEnviron, td->reset, td_setup ))
+			LogError( "Cannot execute reset script %\"s\n", td->reset );
+		sessreg( td, 0, 0, 0 );
+
+		switch ((pid = Fork())) {
+		case 0:
+#if defined(USE_PAM) && defined(HAVE_INITGROUPS)
+			if (restoreGids() && SetUid( curuser, curuid ))
+#else
+			if (SetUser( curuser, curuid, curgid ))
+#endif
+
+			{
+				RemoveUserAuthorization( td );
+#ifdef K5AUTH
+				Krb5Destroy( td->name );
+#endif /* K5AUTH */
+#if !defined(USE_PAM) && !defined(_AIX)
+# ifdef KERBEROS
+				if (krbtkfile[0]) {
+					(void)dest_tkt();
+#  ifndef NO_AFS
+					if (k_hasafs())
+						(void)k_unlog();
+#  endif
+				}
+# endif
+#endif /* !USE_PAM && !_AIX*/
+			}
+			exit( 0 );
+		case -1:
+			LogError( "Cannot clean up session: fork() failed: %m" );
+			break;
+		default:
+			Wait4( pid );
+			break;
+		}
+	}
+
+#ifdef USE_PAM
+	if (removeCreds) {
+# ifdef HAVE_INITGROUPS
+		restoreGids();
+# endif
+		if (removeSession)
+			if ((pretc = pam_close_session( pamh, 0 )) != PAM_SUCCESS)
+				LogError( "pam_close_session() failed: %s\n",
+				          pam_strerror( pamh, pretc ) );
+		if ((pretc = pam_setcred( pamh, PAM_DELETE_CRED )) != PAM_SUCCESS)
+			LogError( "pam_setcred(DELETE_CRED) failed: %s\n",
+			          pam_strerror( pamh, pretc ) );
+		resetGids();
+	}
+	if (pamh) {
+		pam_end( pamh, PAM_SUCCESS );
+		ReInitErrorLog();
+	}
+#endif
+
 	/* make sure the server gets reset after the session is over */
 	if (td->serverPid >= 2) {
 		if (!td->terminateServer && td->resetSignal)
 			TerminateProcess( td->serverPid, td->resetSignal );
 	} else
 		ResetServer( td );
-	if (removeAuth) {
-#ifdef USE_PAM
-		if (pamh) {
-			int pretc;
-			/* shutdown PAM session */
-			if ((pretc = pam_setcred( pamh, PAM_DELETE_CRED )) != PAM_SUCCESS)
-				LogError( "pam_setcred(DELETE_CRED) for %s failed: %s\n",
-				          curuser, pam_strerror( pamh, pretc ) );
-			pam_close_session( pamh, 0 );
-			pam_end( pamh, PAM_SUCCESS );
-			pamh = NULL;
-			ReInitErrorLog();
-		}
-#endif
-
-		if (source( systemEnviron, td->reset, td_setup ))
-			LogError( "Cannot execute reset script %\"s\n", td->reset );
-		sessreg( td, 0, 0, 0 );
-
-		SetUser( curuser, curuid, curgid );
-		RemoveUserAuthorization( td );
-#ifdef K5AUTH
-		Krb5Destroy( td->name );
-#endif /* K5AUTH */
-#if !defined(USE_PAM) && !defined(_AIX)
-# ifdef KERBEROS
-		if (krbtkfile[0]) {
-			(void)dest_tkt();
-#  ifndef NO_AFS
-			if (k_hasafs())
-				(void)k_unlog();
-#  endif
-		}
-# endif
-#endif /* !USE_PAM && !_AIX*/
-#ifdef USE_PAM
-	} else {
-		if (pamh) {
-			pam_end( pamh, PAM_SUCCESS );
-			pamh = NULL;
-			ReInitErrorLog();
-		}
-#endif
-	}
 	Debug( "display %s exiting with status %d\n", td->name, status );
 	exit( status );
 }
@@ -1549,16 +1715,6 @@ ReadDmrc()
 
 	if (!StrApp( &fname, p->pw_dir, "/.dmrc", (char *)0 ))
 		return GE_Error;
-	if ((curdmrc = iniLoad( fname ))) {
-		free( fname );
-		return GE_Ok;
-	}
-
-	if (errno != EACCES) {
-		free( fname );
-		return GE_NoFile;
-	}
-
 	if (pipe( pfd ))
 		return GE_Error;
 	if ((pid = Fork()) < 0) {
